import warnings
from typing import Any, Callable, Dict, Union, overload, TypeVar
import typing_extensions
import reacton
from . import validate_hooks


P = typing_extensions.ParamSpec("P")
FuncT = TypeVar("FuncT", bound=Callable[..., reacton.core.Element])


@overload
def component(obj: None = None, mime_bundle: Dict[str, Any] = ...) -> Callable[[FuncT], FuncT]: ...


@overload
def component(obj: FuncT, mime_bundle: Dict[str, Any] = ...) -> FuncT: ...


@overload
def component(obj: Callable[P, None], mime_bundle: Dict[str, Any] = ...) -> Callable[P, reacton.core.Element]: ...


def component(
    obj: Union[Callable[P, None], FuncT, None] = None, mime_bundle: Dict[str, Any] = reacton.core.mime_bundle_default
) -> Union[Callable[[FuncT], FuncT], FuncT, Callable[P, reacton.core.Element]]:
    def wrapper(obj: Union[Callable[P, None], FuncT]) -> FuncT:
        try:
            validate_hooks.HookValidator(obj).run()
        except Exception as e:
            if not isinstance(e, validate_hooks.HookValidationError):
                # we probably failed because of an unknown reason, but we do not want to break the user's code
                warnings.warn(f"Failed to validate hooks for component {obj.__qualname__}: {e}")
            else:
                raise

        return reacton.component(obj, mime_bundle)  # type: ignore

    if obj is not None:
        return wrapper(obj)
    else:
        return wrapper



from unittest.mock import Mock

import IPython.display

from solara.server import kernel, kernel_context


def test_shell(no_kernel_context):
    ws1 = Mock()
    ws2 = Mock()
    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    kernel1.session.websockets.add(ws1)
    kernel2.session.websockets.add(ws2)
    context1 = kernel_context.VirtualKernelContext(id="1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="2", kernel=kernel2, session_id="session-2")

    with context1:
        IPython.display.display("test1")
        assert ws1.send.call_count == 1
        assert ws2.send.call_count == 0
    with context2:
        IPython.display.display("test1")
        assert ws1.send.call_count == 1
        assert ws2.send.call_count == 1



import hashlib
import os
from pathlib import Path

from pytest import TempPathFactory
import pytest

from solara.server.cdn_helper import get_cdn_url, get_data, get_from_cache, put_in_cache, get_path


def norm(path):
    # this is what starlette does
    return os.path.normpath(os.path.join(*path.split("/")))


path1 = norm("vue-grid-layout@1.0.2/dist/vue-grid-layout.min.js")
hash1 = norm("4bd3c14b1fa124bd9fe4cb5f8a7cbc54")
path2 = norm("@widgetti/vue-grid-layout@2.3.13-alpha.2/dist/vue-grid-layout.umd.js")
hash2 = norm("91c2f41b719978849602e14e17abfb20")


def test_cache(tmp_path_factory):
    base_cache_dir = tmp_path_factory.mktemp("cdn")

    put_in_cache(base_cache_dir, path1, b"test1")
    assert (base_cache_dir / path1).is_file()
    data = get_from_cache(base_cache_dir, path1)
    assert data == b"test1"

    put_in_cache(base_cache_dir, path2, b"test2")
    assert (base_cache_dir / path2).is_file()
    data = get_from_cache(base_cache_dir, path2)
    assert data == b"test2"


def test_cdn_url():
    assert get_cdn_url(path1) == f"https://cdn.jsdelivr.net/npm/{path1}".replace("\\", "/")
    assert get_cdn_url(path2) == f"https://cdn.jsdelivr.net/npm/{path2}".replace("\\", "/")


def test_get_path(tmpdir):
    full_path = get_path(Path(tmpdir), path1)
    assert str(full_path).endswith("vue-grid-layout.min.js")


def test_get_data(tmp_path_factory):
    base_cache_dir = tmp_path_factory.mktemp("cdn")

    # test path1
    data = get_data(base_cache_dir, path1)
    assert hashlib.md5(data).hexdigest() == hash1
    assert (base_cache_dir / path1).is_file()

    assert hashlib.md5((base_cache_dir / path1).read_bytes()).hexdigest() == hash1

    (base_cache_dir / path1).write_bytes(b"test_cached_1")

    data = get_data(base_cache_dir, path1)
    assert data == b"test_cached_1"

    # test path2
    data = get_data(base_cache_dir, path2)
    assert hashlib.md5(data).hexdigest() == hash2
    assert (base_cache_dir / path2).is_file()

    assert hashlib.md5((base_cache_dir / path2).read_bytes()).hexdigest() == hash2

    (base_cache_dir / path2).write_bytes(b"test_cached_2")

    data = get_data(base_cache_dir, path2)
    assert data == b"test_cached_2"


def test_get_data_secure(tmp_path_factory: TempPathFactory):
    # we should never be able to get data from the parent directory

    root_dir = tmp_path_factory.mktemp("root")
    base_cache_dir = root_dir / "cdn"
    base_cache_dir.mkdir()

    (root_dir / "secret").write_bytes(b"not allowed")

    project_dir = base_cache_dir / "project"
    project_dir.mkdir()

    (project_dir / "file").write_bytes(b"a")

    data = get_data(base_cache_dir, "project/file")
    assert data == b"a"
    with pytest.raises(PermissionError):
        get_data(base_cache_dir, "project/../../secret")


def test_redirect(tmp_path_factory):
    base_cache_dir = tmp_path_factory.mktemp("cdn")

    data = get_data(base_cache_dir, "codemirror@5.65.3")

    assert len(data) > 0

    data = get_data(base_cache_dir, "codemirror@5.65.3/lib/codemirror.js")
    assert len(data) > 0


def test_binary(tmp_path_factory):
    base_cache_dir = tmp_path_factory.mktemp("cdn")

    lib = "@widgetti/solara-vuetify-app@0.0.1-alpha.1/dist/037d830416495def72b7881024c14b7b.woff2"
    data = get_data(base_cache_dir, lib)
    assert len(data) == 15436

    assert len((base_cache_dir / lib).read_bytes()) == 15436



# isort: skip_file
from .components import *  # noqa: F401, F403
from .utils import cookies, headers  # noqa: F401, F403
from ..lifecycle import on_kernel_start  # noqa: F401
from ..tasks import task, use_task, Task, TaskResult  # noqa: F401, F403
from ..toestand import computed  # noqa: F401


def __getattr__(name):
    # for backwards compatibility
    from solara.components.cross_filter import (  # noqa: F401
        CrossFilterDataFrame,
        CrossFilterReport,
        CrossFilterSelect,
        CrossFilterSlider,
    )
    from solara.toestand import Reactive, Ref, State  # noqa: F401

    if name == "CrossFilterDataFrame":
        return CrossFilterDataFrame
    elif name == "CrossFilterReport":
        return CrossFilterReport
    elif name == "CrossFilterSelect":
        return CrossFilterSelect
    elif name == "CrossFilterSlider":
        return CrossFilterSlider
    elif name == "Reactive":
        return Reactive
    elif name == "Ref":
        return Ref
    elif name == "State":
        return State
    else:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



import reacton  # noqa: F401
from reacton import ipyvue as rvue  # noqa: F401
from reacton import ipyvuetify as rv  # noqa: F401
from reacton import ipywidgets as rw  # noqa: F401

import solara as sol  # noqa: F401



import ipywidgets as widgets
from concurrent.futures import Future
import playwright.sync_api
from IPython.display import display


def future_trait_change(widget, attribute):
    """Returns a future that will be set when the trait changes."""
    future = Future()  # type: ignore

    def on_change(change):
        # set_result will cause the .result() call below to resume
        future.set_result(change["new"])
        widget.unobserve(on_change, attribute)

    widget.observe(on_change, attribute)
    return future


def test_event_with_polling(solara_test, page_session: playwright.sync_api.Page):
    button = widgets.Button(description="Reset slider")
    slider = widgets.IntSlider(value=42)

    def on_click(button):
        # change the slider value trait when the button is clicked
        # this will be called from the thread the websocket from solara-server
        # is running in, so we can block from the main thread (that pytest is running in)
        slider.value = 0

    button.on_click(on_click)
    display(button)
    # we could display the slider, but it's not necessary for this test
    # since we are only testing if the value changes on the Python side
    # display(slider)
    button_sel = page_session.locator("text=Reset slider")

    # create the future with the attached observer *before* clicking the button
    slider_value = future_trait_change(slider, "value")
    # trigger the click event handler via the frontend, this makes sure that
    # the event handler (on_click) gets executed in a separate thread
    # (the one that the websocket from solara-server is running in)
    button_sel.click()

    # .result() blocks until the value changes or the timeout condition is met.
    # If no value is set, the test will fail due to a TimeoutError
    assert slider_value.result(timeout=2) == 0



"""
# Details
"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "Details"

Page = NoPage

__doc__ += apidoc(solara.Details.f)  # type: ignore



"""
# Tabs

"""

import solara
import solara.lab
from solara.website.utils import apidoc

disabled = solara.reactive(False)
align = solara.reactive("center")
grow = solara.reactive(False)
vertical = solara.reactive(False)
dark = solara.reactive(False)


@solara.component
def Page():
    with solara.Card("Controls"):
        with solara.Row():
            solara.Checkbox(label="Grow", value=grow)
            solara.Checkbox(label="Vertical", value=vertical)
            solara.Checkbox(label="Dark", value=dark)
        if vertical.value:
            align.value = "start"
        else:
            if not grow.value:
                with solara.Row(style={"align-items": "center"}):
                    solara.Text("Align:")
                    solara.ToggleButtonsSingle(value=align, values=["start", "center", "end"], dense=True)
    with solara.lab.Tabs(
        vertical=vertical.value,
        grow=grow.value,
        align=align.value,
        dark=dark.value,
        background_color="primary" if dark.value else None,
        slider_color="secondary" if dark.value else None,
    ):
        with solara.lab.Tab("Tab 1"):
            solara.Markdown("Hello")
        with solara.lab.Tab("Tab 2"):
            solara.Markdown("World")


__doc__ += apidoc(solara.lab.Tabs.f)  # type: ignore



# Disabled, sometimes crashes the webpage.
# import ipycanvas
# import numpy as np
# import solara
# from reacton import ipycanvas as c
# from reacton import ipywidgets as w


# def polygon(canvas, x, y, radius1, radius2, n_points):
#     index = np.arange(n_points)
#     angles = (2 * np.pi / n_points) * index

#     radius = np.where(index % 2, radius1, radius2)
#     v_x = x + np.cos(angles) * radius
#     v_y = y + np.sin(angles) * radius

#     points = np.stack((v_x, v_y), axis=1)

#     canvas.fill_polygon(points)
#     canvas.stroke_polygon(points)


# @solara.component
# def Page():
#     width, height = 800, 800
#     view_count, set_view_count = solara.use_state(0)
#     with w.ViewcountVBox(set_view_count) as main:
#         fill = w.color("#63934e", "fill color")
#         stroke = w.color("#4e6393", "stroke color")
#         line_width = w.slider_int(4, description="line width", min=0, max=30)
#         n_points = w.slider_int(5, "Points", min=1, max=8) * 2
#         radius_inner = w.slider_float(30, "Inner radius", min=0, max=100)
#         radius_outer = w.slider_float(80, "Outer radius", min=0, max=100)

#         def real_drawing():
#             canvas: ipycanvas.Canvas = solara.core.get_widget(canvas_element)

#             with ipycanvas.hold_canvas(canvas):
#                 canvas.clear()
#                 canvas.fill_style = fill
#                 canvas.stroke_style = stroke
#                 canvas.line_width = line_width
#                 radius = width // 3
#                 polygon(canvas, width // 2, height // 2, radius * radius_inner / 100, radius * radius_outer / 100, n_points)

#         solara.use_effect(real_drawing, [fill, stroke, line_width, n_points, view_count, radius_inner, radius_outer])
#         canvas_element = c.Canvas(width=width, height=height)

#     return main



from pathlib import Path
import pytest
import solara

try:
    import vaex
except ImportError:
    vaex = None
import polars as pl
import pandas as pd

from solara.components.datatable import DataTable, DataTableWidget


HERE = Path(__file__).parent

titanic_url = HERE.parent / "titanic.csv"
df_pandas = pd.read_csv(titanic_url)
# convert columns to numpy-backed types for polars conversion without pyarrow
# df_pandas.index = df_pandas.index.astype('int64')
for col in df_pandas.columns:
    # Check specifically for pandas nullable dtypes, checking for attribute existence for compatibility
    if hasattr(pd, "Int64Dtype") and isinstance(df_pandas[col].dtype, pd.Int64Dtype):
        # Fill NaN and convert to numpy's int64
        df_pandas[col] = df_pandas[col].fillna(0).astype("int64")
    elif hasattr(pd, "Float64Dtype") and isinstance(df_pandas[col].dtype, pd.Float64Dtype):
        # Fill NaN and convert to numpy's float64
        df_pandas[col] = df_pandas[col].fillna(0.0).astype("float64")
    elif df_pandas[col].dtype == "object":
        # Try converting object columns to string
        try:
            df_pandas[col] = df_pandas[col].astype(str)
            # You might need more specific NaN handling here depending on the column
            # For now, converting NaN to the string 'nan'
            df_pandas[col] = df_pandas[col].fillna("nan")
        except Exception as e:
            print(f"Could not convert object column {col} to string: {e}")
            # Handle specific conversion errors if necessary

if vaex is not None:
    df_vaex = vaex.from_pandas(df_pandas)
df_polars = pl.from_pandas(df_pandas)


@pytest.mark.parametrize("df", [df_vaex, df_pandas, df_polars] if vaex is not None else [df_pandas, df_polars])
def test_render(df):
    @solara.component
    def Test():
        return DataTable(df)

    widget, rc = solara.render_fixed(Test(), handle_error=False)
    assert isinstance(widget, DataTableWidget)
    assert len(widget.items) == 20



import ipyvuetify as v
import pytest

import solara
import solara.lab


def test_tabs_no_children():
    @solara.component
    def Test():
        with solara.lab.Tabs():
            solara.lab.Tab("Tab 1")
            solara.lab.Tab("Tab 2")

    box, rc = solara.render(Test(), handle_error=False)
    assert len(rc.find(v.Tab)) == 2
    assert len(rc.find(v.Tab)) == 2
    assert rc.find(v.Tabs).widget.v_model == 0


@pytest.mark.parametrize("lazy", [True, False])
def test_tabs_basics(lazy):
    @solara.component
    def Test():
        index, set_index = solara.use_state(0)
        with solara.lab.Tabs(value=index, on_value=set_index, lazy=lazy):
            with solara.lab.Tab("Tab 1"):
                solara.SliderInt("Slider 1", value=1)
            with solara.lab.Tab("Tab 2"):
                solara.InputText("Input 2", value="2")

    box, rc = solara.render(Test(), handle_error=False)
    assert len(rc.find(v.Tab)) == 2
    assert len(rc.find(v.TabsItems)) == 1
    if lazy:
        assert len(rc.find(v.TabItem)) == 2
        assert len(rc.find(v.SkeletonLoader)) == 0
        assert len(rc.find(v.Slider)) == 1
        assert len(rc.find(v.TextField)) == 0
    else:
        assert len(rc.find(v.TabItem)) == 2
        assert len(rc.find(v.SkeletonLoader)) == 0
        assert len(rc.find(v.SkeletonLoader)) == 0
        assert len(rc.find(v.Slider)) == 1
        assert len(rc.find(v.TextField)) == 1
    assert rc.find(v.Tabs).widget.v_model == 0



from PyInstaller.utils.hooks import collect_data_files, copy_metadata, collect_submodules

hiddenimports = collect_submodules("ipyreact")
datas = collect_data_files("ipyreact")  # codespell:ignore datas
datas += copy_metadata("ipyreact")  # codespell:ignore datas



import solara
import solara.lab

value_reactive = solara.reactive(1.0)


@solara.lab.computed
def computed_reactive():
    return value_reactive.value + 1.0


@solara.component
def Page():
    solara.FloatSlider("test", value=value_reactive)
    solara.InputFloat("test", value=computed_reactive.value)



import solara


@solara.component
def Page1():
    solara.Text("Page 1")


@solara.component
def Page2():
    solara.Text("Page 2")


@solara.component
def Layout(children):
    with solara.AppLayout():
        with solara.Column():
            solara.Text("Custom layout")
            solara.display(*children)


routes = [
    solara.Route("/", component=Page1, layout=Layout),
    solara.Route("page2", component=Page2),
]



import re
import sys
import threading
from typing import MutableMapping

from .types import ObservableMutableMapping


def _in_solara_server():
    return "solara.server" in sys.modules


class ConnectionScope(ObservableMutableMapping):
    def __init__(self, name="connection"):
        super().__init__()
        self._global_dict = {}
        self.name = name
        self.lock = threading.Lock()

    def _get_dict(self) -> MutableMapping:
        if _in_solara_server():
            import solara.server.kernel_context

            context = solara.server.kernel_context.get_current_context()
            if self.name not in context.user_dicts:
                with self.lock:
                    if self.name not in context.user_dicts:
                        context.user_dicts[self.name] = {}
            return context.user_dicts[self.name]
        else:
            return self._global_dict


class ObservableDict(ObservableMutableMapping):
    def __init__(self):
        super().__init__()
        self._global_dict = {}

    def _get_dict(self) -> MutableMapping:
        return self._global_dict


worker = ObservableDict()
connection = ConnectionScope()


def get_kernel_id(ipython_fallback=True) -> str:
    """Returns the kernel id, a unique string for each virtual kernel.

    See [Understanding solara server](/docs/understanding/solara-server) for understanding the concept of virtual kernels
    and their lifetime.

    This unique ID can be useful to to implement storing state, scoped to a kernel. See [the scope example](/examples/general/custom_storage) for an example.

    If `ipython_fallback` is `True` (default), this function will also work in IPython notebooks, where it will return the IPython kernel id.

    """
    import solara.server.kernel_context

    try:
        context = solara.server.kernel_context.get_current_context()
    except RuntimeError as e:
        if not ipython_fallback:
            raise
        import IPython

        ipython = IPython.get_ipython()
        if not ipython or not hasattr(ipython, "kernel"):
            raise RuntimeError("Not in a kernel") from e
        kernel = ipython.kernel
        regex = r"[\\/]kernel-([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})\.json$"
        connection_file = kernel.config["IPKernelApp"]["connection_file"]
        return re.compile(regex).search(connection_file).group(1)  # type: ignore
    return context.id


def get_session_id() -> str:
    """Returns the session id, which is stored using a browser cookie.

    See [Understanding solara server](/docs/understanding/solara-server#session) for more information about the Solara sessions.

    This unique ID can be useful to to implement storing state, scoped to a browser session. See [the scope example](/examples/general/custom_storage)
    for an example.
    """
    import solara.server.kernel_context

    context = solara.server.kernel_context.get_current_context()
    return context.session_id



import operator
from typing import Any, Dict, List, Optional, cast

import ipyvuetify
import reacton.ipyvuetify as v
import solara
import traitlets
from solara import CellAction, ColumnAction

from ..lab.hooks.dataframe import use_df_column_names
from ..lab.utils.dataframe import (
    df_filter_values,
    df_py_types,
    df_range,
    df_value_count,
)

# to avoid confusing vuetify about selecting 'None' and nothing
magic_value_missing = "__missing_value__"

# first selector is needed in jupyter, which needs to be more specific
# second selector in solara, where there is no need to be more specific (css load order?) and there is not .vuetify-styles
css_message = """
.vuetify-styles .solara-cross-filter-select .v-messages,
.solara-cross-filter-select .v-messages {
    color: #fb8c00;
}
"""


class Select(ipyvuetify.VuetifyTemplate):
    template_file = (__file__, "select.vue")

    value = traitlets.Any().tag(sync=True)
    label = traitlets.Unicode().tag(sync=True)
    clearable = traitlets.Bool().tag(sync=True)
    return_object = traitlets.Bool().tag(sync=True)
    items = traitlets.List(cast(List[Dict[str, Any]], [])).tag(sync=True)
    filtered = traitlets.Bool().tag(sync=True)
    count = traitlets.Int().tag(sync=True)
    multiple = traitlets.Bool().tag(sync=True)
    messages = traitlets.Unicode().tag(sync=True)


@solara.component
def CrossFilterSelect(
    df,
    column: str,
    max_unique: int = 100,
    multiple: bool = False,
    invert=False,
    configurable=True,
    classes: List[str] = [],
):
    """A Select widget that will cross filter a DataFrame.

    See [use_cross_filter](/documentation/api/hooks/use_cross_filter) for more information about how to use cross filtering.

    ## Arguments

    - `df`: The DataFrame to filter.
    - `column`: The column to filter on.
    - `max_unique`: The maximum number of unique values to show in the dropdown.
    - `multiple`: Whether to allow multiple values to be selected.
    - `invert`: Whether to invert the selection.
    - `configurable`: Whether to show the configuration button.
    - `classes`: Additional CSS classes to add to the main widget.

    """
    import pandas as pd

    filter, set_filter = solara.use_cross_filter(id(df), "filter-dropdown")
    filter_values, set_filter_values = solara.use_state(cast(List[Any], []))
    column, set_column = solara.use_state_or_update(column)
    invert, set_invert = solara.use_state_or_update(invert)
    multiple, set_multiple = solara.use_state_or_update(multiple)

    dff = df
    if filter is not None:
        dff = df[filter]

    value_counts = df_value_count(df, column, limit=max_unique + 1)
    value_counts.rename({"count": "count_max"}, axis=1, inplace=True)

    value_counts_filtered = df_value_count(dff, column, limit=max_unique + 1)
    value_counts = value_counts.merge(value_counts_filtered, how="left", on="value")
    value_counts["count"] = value_counts["count"].fillna(0)
    value_counts["exists"] = value_counts["count"] > 0
    value_counts.sort_values(["exists", "value"], ascending=[False, True], inplace=True)

    columns = use_df_column_names(df)

    def set_values_and_filter(values):
        if values is None:
            set_filter_values([])
            return

        if multiple:
            set_filter_values([value["value"] for value in values])
        else:
            set_filter_values([values["value"]])

    def reset():
        set_filter_values([])

    solara.use_memo(reset, dependencies=[column])

    def update_filter():
        if len(filter_values) == 0:
            set_filter(None)
        else:
            filter_values_without_magic = [None if k == magic_value_missing else k for k in filter_values]
            filter = df_filter_values(df, column, filter_values_without_magic, invert=invert)
            set_filter(filter)

    solara.use_memo(update_filter, dependencies=[filter_values, invert])

    items = [
        {
            "value": magic_value_missing if pd.isna(k.value) else k.value,
            "text": str(k.value) if not pd.isna(k.value) else "NA",
            "count": k.count,
            "count_max": k.count_max,
        }
        for k in value_counts.itertuples()
    ]
    value: Any = None
    if not multiple:
        value = {"value": filter_values[0]} if len(filter_values) > 0 else None
    else:
        value = [{"value": k} for k in filter_values]
    # TODO: reacton bug, we cannot add this under any component context manager
    # this gives an error, probably because the button is added twice
    with v.Btn(v_on="x.on", icon=True) as btn:
        v.Icon(children=["mdi-settings"])
    with solara.VBox(classes=classes) as main:
        solara.Style(css_message)
        with solara.HBox(align_items="baseline"):
            label = f"Select values in {column} having values:" if not invert else f"Drop values in {column} having values:"
            Select.element(
                value=value,
                items=items,
                on_value=set_values_and_filter,
                label=label,
                clearable=True,
                return_object=True,
                multiple=multiple,
                filtered=filter is not None,
                count=len(dff),
                messages=f"Too many unique values, will only show the first {max_unique}" if len(value_counts) > max_unique else "",
                class_="solara-cross-filter-select",
            )
            if configurable:
                with v.Menu(v_slots=[{"name": "activator", "variable": "x", "children": btn}], close_on_content_click=False):
                    with v.Sheet():
                        with v.Container(py_0=True, px_3=True, ma_0=True):
                            with v.Row():
                                with v.Col():
                                    v.Select(v_model=column, items=columns, on_v_model=set_column, label="Choose column")
                                    v.Switch(v_model=invert, on_v_model=set_invert, label="Invert filter")
                                    v.Switch(v_model=multiple, on_v_model=set_multiple, label="Select multiple")

    return main


@solara.component
def CrossFilterReport(df, classes: List[str] = []):
    """Shows a report of the current cross filter state.

    Shows number of rows filtered, and the total number of rows.

    See [use_cross_filter](/documentation/api/hooks/use_cross_filter) for more information about how to use cross filtering.

    ## Arguments

    - `df`: The DataFrame where the filter is applied to.
    - `classes`: Additional CSS classes to add to the main widget.

    """
    filter, set_filter = solara.use_cross_filter(id(df), "summary")
    dff = df
    filtered = False
    if filter is not None:
        filtered = True
        dff = df[filter]
    progress = len(dff) / len(df) * 100
    with solara.VBox(classes=classes) as main:
        with solara.HBox(align_items="center"):
            icon = "mdi-filter"
            v.Icon(children=[icon], style_="opacity: 0.1" if not filtered else "")
            if filtered:
                summary = f"{len(dff):,} / {len(df):,}"
            else:
                summary = f"{len(dff):,}"
            v.Html(tag="h3", children=[summary], style_="display: inline")
        # always add a progress bar to make sure the layout is the same
        if filtered:
            v.ProgressLinear(value=progress).key("visible")
        else:
            v.ProgressLinear(value=0, style_="visibility: hidden").key("hidden")

    return main


@solara.component
def CrossFilterSlider(
    df,
    column: str,
    invert=False,
    enable: bool = True,
    mode: str = "==",
    configurable=True,
):
    """A Slider widget that will cross filter a DataFrame.

    See [use_cross_filter](/documentation/api/hooks/use_cross_filter) for more information about how to use cross filtering.

    ## Arguments

    - `df`: The DataFrame to filter.
    - `column`: The column to filter on.
    - `invert`: If True, the filter will be inverted.
    - `enable`: If False, the filter will be disabled.
    - `mode`: The mode to use for filtering. Can be one of `==`, `>=`, `<=`, `>`, `<`.
    - `configurable`: Whether to show a configuration button.

    """
    filter, set_filter = solara.use_cross_filter(id(df), "filter-slider")
    filter_value, set_filter_value = solara.use_state(cast(Optional[float], None))
    column, set_column = solara.use_state_or_update(column)
    invert, set_invert = solara.use_state_or_update(invert)
    enable, set_enable = solara.use_state_or_update(enable)
    mode, set_mode = solara.use_state_or_update(mode)

    # TODO: should we use the filter for min/max?
    # dff = df
    # if filter is not None:
    #     dff = df[filter]

    vmin, vmax = df_range(df, column)

    columns = use_df_column_names(df)
    py_types = df_py_types(df)
    columns_numeric = [c for c in columns if py_types[c] in [int, float]]

    def reset():
        set_filter_value(vmin)

    solara.use_memo(reset, dependencies=[column])

    def update_filter():
        if not enable or filter_value is None:
            set_filter(None)
        else:
            operator_map = {
                "==": operator.eq,
                ">=": operator.ge,
                "<=": operator.le,
                ">": operator.gt,
                "<": operator.lt,
                "!=": operator.ne,
            }
            filter = operator_map[mode](df[column], filter_value)
            if invert:
                filter = ~filter
            set_filter(filter)

    solara.use_memo(update_filter, dependencies=[filter_value, invert, enable, mode])

    # TODO: reacton bug, see CrossFilterSelect
    with v.Btn(v_on="x.on", icon=True) as btn:
        v.Icon(children=["mdi-settings"])

    with solara.VBox() as main:
        with solara.HBox(align_items="center"):
            label = f"Select {column} {mode} " if not invert else f"Drop {column} {mode} "

            if py_types[column] is int:
                if isinstance(filter_value, int):
                    solara.SliderInt(label=label, value=filter_value, min=vmin, max=vmax, on_value=set_filter_value, disabled=not enable, thumb_label=False)
                else:
                    solara.Error(f"Filter value is not an integer type, but {type(filter_value)} (value = {filter_value})")
                if filter_value is not None:
                    solara.Text(f"{filter_value:,}")
            elif py_types[column] is float:
                if isinstance(filter_value, float):
                    solara.SliderFloat(label=label, value=filter_value, min=vmin, max=vmax, on_value=set_filter_value, disabled=not enable, thumb_label=False)
                else:
                    solara.Error(f"Filter value is not an floating point type, but {type(filter_value)} (value = {filter_value})")
                if filter_value is not None:
                    solara.Text(f"{filter_value:,}")
            else:
                solara.Warning(f"{py_types[column]} not supported for Slider")

            if configurable:
                with v.Menu(v_slots=[{"name": "activator", "variable": "x", "children": btn}], close_on_content_click=False):
                    with v.Sheet():
                        with v.Container(py_0=True, px_3=True, ma_0=True):
                            with v.Row():
                                with v.Col():
                                    columns_numeric = [c for c in columns if py_types[c] in [int, float]]
                                    v.Select(v_model=column, items=columns_numeric, on_v_model=set_column, label="Choose column")
                                    v.Switch(v_model=invert, on_v_model=set_invert, label="Invert filter")
                                    v.Switch(v_model=enable, on_v_model=set_enable, label="Enable filter")
                                    with solara.ToggleButtonsSingle(value=mode, on_value=set_mode):  # type: ignore
                                        solara.Button(icon_name="mdi-code-equal", icon=True, value="==")
                                        solara.Button(icon_name="mdi-code-not-equal", icon=True, value="!=")
                                        solara.Button(icon_name="mdi-code-less-than", icon=True, value="<")
                                        solara.Button(icon_name="mdi-code-less-than-or-equal", icon=True, value="<=")
                                        solara.Button(icon_name="mdi-code-greater-than", icon=True, value=">")
                                        solara.Button(icon_name="mdi-code-greater-than-or-equal", icon=True, value=">=")

    return main


@solara.component
def CrossFilterDataFrame(df, items_per_page=20, column_actions: List[ColumnAction] = [], cell_actions: List[CellAction] = [], scrollable=False):
    """Display a DataFrame with filters applied from the cross filter.

    This component wraps [DataFrame](/documentation/components/data/dataframe).

    See [use_cross_filter](/documentation/api/hooks/use_cross_filter) for more information about how to use cross filtering.

    # Arguments

     * `df` - a Pandas dataframe.
     * `column_actions` - Triggered via clicking on the triple dot icon on the headers (visible when hovering).
     * `cell_actions` -  Triggered via clicking on the triple dot icon in the cell (visible when hovering).

    """
    dff = df
    filter, set_filter = solara.use_cross_filter(id(df), "dataframe")
    if filter is not None:
        dff = df[filter]
    return solara.DataFrame(dff, items_per_page=items_per_page, scrollable=scrollable, column_actions=column_actions, cell_actions=cell_actions)



"""# Info

Solara has 4 types of alerts:

 * [Success](/documentation/components/status/success)
 * Info (this page)
 * [Warning](/documentation/components/status/warning)
 * [Error](/documentation/components/status/error)



"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    icon = solara.use_reactive(True)
    dense = solara.use_reactive(False)
    outlined = solara.use_reactive(True)
    text = solara.use_reactive(True)

    with solara.GridFixed(4):
        solara.Checkbox(label="Use icon", value=icon)
        solara.Checkbox(label="Show dense", value=dense)
        solara.Checkbox(label="Show as text", value=text)
        solara.Checkbox(label="Show outlined", value=outlined)
    solara.Info(
        f"This is solara.Info(label='...', text={text.value}, dense={dense.value}, outlined={outlined.value}, icon={icon.value})",
        text=text.value,
        dense=dense.value,
        outlined=outlined.value,
        icon=icon.value,
    )


__doc__ += apidoc(solara.Info.f)  # type: ignore



from typing import Callable, Dict, List, Optional, Tuple, Union, cast

import reacton
import reacton.core
import reacton.ipyvuetify as v
import reacton.utils
from reacton.core import Element

import solara
import solara.lab

from . import title as t


@solara.component
def AppIcon(open=False, on_click=None, **kwargs):
    def click(*ignore):
        on_click()

    icon = v.AppBarNavIcon(**kwargs)
    v.use_event(icon, "click", click)
    return icon


should_use_embed = solara.create_context(False)
PortalElements = Dict[str, List[Tuple[int, Element]]]


def _set_sidebar_default(updater: Callable[[PortalElements], PortalElements]):
    pass


class ElementPortal:
    def __init__(self):
        self.context = solara.create_context(_set_sidebar_default)

    # TODO: can we generalize the use of 'portals' ? (i.e. transporting elements from one place to another)
    def use_portal(self) -> List[Element]:
        portal_elements, set_portal_elements = solara.use_state(cast(PortalElements, {}))
        self.context.provide(set_portal_elements)  # type: ignore

        portal_elements_flat: List[Tuple[int, Element]] = []
        for uuid, value in portal_elements.items():
            portal_elements_flat.extend(value)
        portal_elements_flat.sort(key=lambda x: x[0])
        return [e[1] for e in portal_elements_flat]

    def use_portal_add(self, children: List[Element], offset: int):
        key = solara.use_unique_key(prefix="portal-")
        set_portal_elements = solara.use_context(self.context)
        values: List[Tuple[int, Element]] = []
        for i, child in enumerate(children):
            values.append((offset + i, child))

        # updates we do when children/offset changes
        def add():
            # we use the update function method, to avoid stale data
            def update_dict(portal_elements):
                portal_elements_updated = portal_elements.copy()
                portal_elements_updated[key] = values
                return portal_elements_updated

            set_portal_elements(update_dict)

        solara.use_effect(add, [values])

        # cleanup we only need to do after component removal
        def add_cleanup():
            def cleanup():
                def without(portal_elements):
                    portal_elements_restored = portal_elements.copy()
                    portal_elements_restored.pop(key, None)
                    return portal_elements_restored

                set_portal_elements(without)

            return cleanup

        solara.use_effect(add_cleanup, [])


sidebar_portal = ElementPortal()
appbar_portal = ElementPortal()
apptitle_portal = ElementPortal()


@solara.component
def AppBar(children=[]):
    """Puts its children in the app bar of the AppLayout (or any layout that supports it).

    This component does not need to be a direct child of the AppLayout, it can be at any level in your component tree.

    If a [Tabs](/documentation/components/lab/tabs) component is used as direct child of the app bar, it will be shown under the app bar.

    ## Example showing an app bar
    ```solara
    import solara

    @solara.component
    def Page():
        logged_in, set_logged_in = solara.use_state(False)
        def toggle_login():
            set_logged_in(not logged_in)

        with solara.AppBar():
            icon_name = "mdi-logout" if logged_in else "mdi-login"
            solara.Button(icon_name=icon_name , on_click=toggle_login, icon=True)
        with solara.Column():
            if logged_in:
                solara.Info("You are logged in")
            else:
                solara.Error("You are logged out")
    ```
    """
    # TODO: generalize this, this is also used in title
    level = 0
    rc = reacton.core.get_render_context()
    context = rc.context
    while context and context.parent:
        level += 1
        context = context.parent
    offset = 2**level
    appbar_portal.use_portal_add(children, offset)

    return solara.Div(style="display; none")


@solara.component
def AppBarTitle(children=[]):
    """Puts its children in the title section of the AppBar (or any layout that supports it).

    This component does not need to be a direct child of the AppBar, it can be at any level in your component tree.

    ## Example

    ```solara
    import solara

    @solara.component
    def Page():
        with solara.AppBarTitle():
            solara.Text("Hi there")
            solara.Button("Click me", outlined=True, classes=["mx-2"])
    ```
    """
    level = 0
    rc = reacton.core.get_render_context()
    context = rc.context
    while context and context.parent:
        level += 1
        context = context.parent
    offset = 2**level
    apptitle_portal.use_portal_add(children, offset)

    return solara.Div(style="display; none")


@solara.component
def Sidebar(children=[]):
    """Puts its children in the sidebar of the AppLayout (or any layout that supports it).
    This component does not need to be a direct child of the AppLayout, it can be at any level in your component tree.

    On the solara.dev website and in the Jupyter notebook, the sidebar is shown in a dialog instead (embedded mode)

    ## Example showing a sidebar (embedded mode)
    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Column() as main:
            with solara.Sidebar():
                solara.Markdown("## I am in the sidebar")
                solara.SliderInt(label="Ideal for placing controls")
            solara.Info("I'm in the main content area, put your main content here")
        return main
    ```


    """
    # TODO: generalize this, this is also used in title
    level = 0
    rc = reacton.core.get_render_context()
    context = rc.context
    while context and context.parent:
        level += 1
        context = context.parent
    offset = 2**level
    sidebar_portal.use_portal_add(children, offset)

    return solara.Div(style="display; none")


@solara.component
def AppLayout(
    children=[],
    sidebar_open=True,
    title=None,
    navigation=True,
    toolbar_dark=True,
    color: Optional[str] = "primary",
    classes: List[str] = [],
    style: Optional[Union[str, Dict[str, str]]] = None,
):
    """The default layout for Solara apps. It consists of an toolbar bar, a sidebar and a main content area.

     * The title of the app is set using the [Title](/documentation/components/page/title) component.
     * The sidebar content is set using the [Sidebar](/documentation/components/layout/sidebar) component.
     * The content is set by the `Page` component provided by the user.

    This component is usually not used directly, but rather through via the [Layout system](/documentation/advanced/howto/layout).

    The sidebar is only added when the AppLayout has more than one child.

    ```python
    with AppLayout(title="My App"):
        with v.Card():
            ...  # sidebar content
        with v.Card():
            ...  # main content
    ```

    # Arguments

     * `children`: The children of the AppLayout. The first child is used as the sidebar content, the rest as the main content.
     * `sidebar_open`: Whether the sidebar is open or not.
     * `title`: The title of the app shown in the app bar, can also be set using the [Title](/documentation/components/page/title) component.
     * `toolbar_dark`: Whether the toolbar should be dark or not.
     * `navigation`: Whether the navigation tabs based on routing should be shown.
     * `color`: The color of the toolbar.
     * `classes`: List of CSS classes to apply to the direct parent of the childred.
     * `style`: CSS style to apply to the direct parent of the children. If style is None we use a default style of "height: 100%; overflow: auto;"
        and add 12px of padding when the sidebar of titlebar is visible. This will make sure your app gets scrollbars when need.
    """
    route, routes = solara.use_route()
    paths = [solara.resolve_path(r, level=0) for r in routes]
    location = solara.use_context(solara.routing._location_context)
    embedded_mode = solara.use_context(should_use_embed)
    fullscreen, set_fullscreen = solara.use_state(False)
    # we cannot nest AppLayouts, so we can use the context to set the embedded mode
    should_use_embed.provide(True)
    index = routes.index(route) if route else None

    sidebar_open, set_sidebar_open = solara.use_state_or_update(sidebar_open)
    # remove the appbar from the children
    children_without_portal_sources = [c for c in children if c.component != AppBar]
    use_drawer = len(children_without_portal_sources) > 1
    children_content = children
    children_sidebar = []
    if use_drawer:
        child_sidebar = children_without_portal_sources.pop(0)
        children_sidebar = [child_sidebar]
        children_content = [c for c in children if c is not child_sidebar]
    children_sidebar = children_sidebar + sidebar_portal.use_portal()
    children_appbar = appbar_portal.use_portal()
    if children_sidebar:
        use_drawer = True
    title = t.use_title_get() or title
    children_appbartitle = apptitle_portal.use_portal()
    v_slots = []

    tabs_element = None
    for child_appbar in children_appbar.copy():
        if child_appbar.component == solara.lab.Tabs:
            if tabs_element is not None:
                raise ValueError("Only one Tabs component is allowed in the AppBar")
            tabs_element = child_appbar
            children_appbar.remove(tabs_element)

    show_app_bar = (title and (len(routes) > 1 and navigation)) or bool(children_appbar) or bool(use_drawer) or bool(children_appbartitle) or bool(tabs_element)

    if style is None:
        style = {"height": "100%", "max-height": "100%", "overflow": "auto"}
        # if style is None, we choose a default style based on whether we are seeing the appbar, etc
        if show_app_bar or children_sidebar or len(children) != 1:
            style["padding"] = "12px"

    def set_path(index):
        path = paths[index]
        location.pathname = path

    if (tabs_element is None) and routes and navigation and (len(routes) > 1):
        with solara.lab.Tabs(value=index, on_value=set_path, align="center") as tabs_element:
            for route in routes:
                name = route.path if route.path != "/" else "Home"
                solara.lab.Tab(name)
        # with v.Tabs(v_model=index, on_v_model=set_path, centered=True) as tabs:
        #     for route in routes:
        #         name = route.path if route.path != "/" else "Home"
        #         v.Tab(children=[name])
    if tabs_element is not None and navigation:
        v_slots = [{"name": "extension", "children": tabs_element}]
    if embedded_mode and not fullscreen:
        # this version doesn't need to run fullscreen
        # also ideal in jupyter notebooks
        with v.Html(tag="div") as main:
            if show_app_bar or use_drawer:
                with v.AppBar(color=color, dark=toolbar_dark, v_slots=v_slots):
                    if use_drawer:
                        icon = AppIcon(sidebar_open, on_click=lambda: set_sidebar_open(not sidebar_open), v_on="x.on")
                        with v.Menu(
                            offset_y=True,
                            nudge_left="50px",
                            left=True,
                            v_slots=[{"name": "activator", "variable": "x", "children": [icon]}],
                            close_on_content_click=False,
                        ):
                            pass
                            v.Sheet(children=children_sidebar, style_="padding: 12px; min-width: 400px")
                    if title or children_appbartitle:
                        v.ToolbarTitle(children=children_appbartitle or [title])
                    v.Spacer()
                    for i, child in enumerate(children_appbar):
                        # if the user already provided a key, don't override it
                        if child._key is None:
                            solara.display(child.key(f"app-layout-appbar-user-child-{i}"))
                        else:
                            solara.display(child)
                    solara.Button(icon_name="mdi-fullscreen", on_click=lambda: set_fullscreen(True), icon=True, dark=False)
            with v.Row(no_gutters=False, class_="solara-content-main"):
                v.Col(cols=12, children=children_content)
    else:
        # this limits the height of the app to the height of the screen
        # and further down we use overflow: auto to add scrollbars to the main content
        # the navigation drawer adds it own scrollbars
        # NOTE: while developing this we added overflow: hidden, but this does not seem
        # to be necessary anymore
        with v.Html(tag="div", style_="height: 100vh") as main:
            with solara.HBox():
                if use_drawer:
                    with v.NavigationDrawer(
                        width="min-content",
                        v_model=sidebar_open,
                        on_v_model=set_sidebar_open,
                        style_="min-width: 400px; max-width: 600px",
                        clipped=True,
                        app=True,
                        # disable_resize_watcher=True,
                        disable_route_watcher=True,
                        mobile_break_point="960",
                        class_="solara-content-main",
                    ):
                        if not show_app_bar:
                            AppIcon(sidebar_open, on_click=lambda: set_sidebar_open(not sidebar_open))
                        v.Html(tag="div", children=children_sidebar, style_="padding: 12px; height: 100%").meta(ref="sidebar-content")
            if show_app_bar:
                # if hide_on_scroll is True, and we have a little bit of scrolling, vuetify seems to act strangely
                # when scrolling (on @mariobuikhuizen/vuetify v2.2.26-rc.0
                with v.AppBar(color=color, dark=toolbar_dark, app=True, clipped_left=True, hide_on_scroll=False, v_slots=v_slots).key("app-layout-appbar"):
                    if use_drawer:
                        AppIcon(sidebar_open, on_click=lambda: set_sidebar_open(not sidebar_open))
                    if title or children_appbartitle:
                        v.ToolbarTitle(children=children_appbartitle or [title])
                    v.Spacer()
                    for i, child in enumerate(children_appbar):
                        # if the user already provided a key, don't override it
                        if child._key is None:
                            solara.display(child.key(f"app-layout-appbar-user-child-{i}"))
                        else:
                            solara.display(child)
                    if fullscreen:
                        solara.Button(icon_name="mdi-fullscreen-exit", on_click=lambda: set_fullscreen(False), icon=True, dark=False)
            # in vue2 is was v-content, in vue3 it is v-main
            MainComponent = v.Main if solara.util.ipyvuetify_major_version == 3 else v.Content  # type: ignore
            with MainComponent(class_="solara-content-main", style_="height: 100%;").key("app-layout-content"):
                # make sure the scrollbar does no go under the appbar by adding overflow: auto
                # to a child of content, because content has padding-top: 64px (set by vuetify)
                # the padding: 12px is needed for backward compatibility with the previously used
                # v.Col which has this by default. If we do not use this, a solara.Column will
                # use a margin: -12px which will make a horizontal scrollbar appear
                solara.Div(style=style, classes=classes, children=children_content)
        if fullscreen:
            with v.Dialog(v_model=True, children=[], fullscreen=True, hide_overlay=True, persistent=True, no_click_animation=True) as dialog:
                v.Sheet(class_="overflow-y-auto overflow-x-auto", children=[main])
                pass
            return dialog
    return main


@solara.component
def _AppLayoutEmbed(children=[], sidebar_open=True, title=None):
    """Forces the embed more for a AppLayout. This is used by default in Jupyter."""
    should_use_embed.provide(True)

    if solara.checks.should_perform_jupyter_check():
        children = [solara.Column(children=children + [solara.checks.JupyterCheck()])]

    def once():
        import solara.server.telemetry

        solara.server.telemetry.jupyter_start()

    solara.use_effect(once, [])
    return AppLayout(children=children, sidebar_open=sidebar_open, title=title)


reacton.core.jupyter_decorator_components.append(_AppLayoutEmbed)



# import sys

import ipywidgets as widgets

from solara.server import kernel, kernel_context

# import pytest


# temporary disabled
# # with python 3.6 we don't use the comm package
# @pytest.mark.skipif(sys.version_info < (3, 7, 0), reason="ipykernel version too low")
# def test_widget_error_message_outside_context(no_kernel_context):
#     from ipyvuetify.Themes import theme

#     theme.get_state()
#     kernel_shared = kernel.Kernel()
#     context1 = kernel_context.VirtualKernelContext(id="1", kernel=kernel_shared, session_id="session-1")
#     with pytest.raises(RuntimeError):
#         with context1:
#             assert theme.model_id


def test_widget_dict(no_kernel_context):
    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    context1 = kernel_context.VirtualKernelContext(id="1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="2", kernel=kernel2, session_id="session-2")

    with context1:
        btn1 = widgets.Button(description="context1")
    with context2:
        btn2 = widgets.Button(description="context2")

    assert btn1.model_id in context1.widgets
    assert btn1.model_id not in context2.widgets
    assert btn2.model_id in context2.widgets
    assert btn2.model_id not in context1.widgets

    del btn1
    del btn2
    context1.close()
    context2.close()



from pathlib import Path

from ipyvuetify import Btn

import solara
from solara.components.file_download import FileDownloadWidget


def test_download(tmpdir: Path):
    _, rc = solara.render(solara.FileDownload(data="test content"), handle_error=False)
    assert rc.find(FileDownloadWidget).widget.filename == "solara-download.dat"
    assert rc.find(FileDownloadWidget).widget.bytes is None
    # lazily upload the data
    rc.find(FileDownloadWidget).widget.request_download = True
    rc.find(FileDownloadWidget, bytes=b"test content").wait_for()
    assert rc.find(Btn).widget.children[1] == "Download: solara-download.dat"

    # change the data
    rc.render(solara.FileDownload(data="test content2"))
    # it should 'reset'
    rc.find(FileDownloadWidget).widget.request_download = False
    assert rc.find(FileDownloadWidget).widget.bytes is None
    rc.find(FileDownloadWidget).widget.request_download = True
    rc.find(FileDownloadWidget, bytes=b"test content2").wait_for()

    filename = tmpdir / "test.txt"
    filename.write_text("test content3", encoding="utf-8")
    rc.render(solara.FileDownload(data=filename.open("rb")))
    assert rc.find(Btn).widget.children[1] == "Download: test.txt"
    rc.find(FileDownloadWidget).widget.request_download = True
    rc.find(FileDownloadWidget, bytes=b"test content3").wait_for()

    lazy_read_calls = 0

    def lazy_read():
        nonlocal lazy_read_calls
        lazy_read_calls += 1
        return b"lazy read"

    rc.render(solara.FileDownload(data=lazy_read))
    assert rc.find(Btn).widget.children[1] == "Download: solara-download.dat"
    assert lazy_read_calls == 0
    rc.find(FileDownloadWidget).widget.request_download = True
    assert lazy_read_calls == 1
    rc.find(FileDownloadWidget, bytes=b"lazy read").wait_for()



"""# Task"""

import solara
import solara.autorouting
import solara.lab
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "Task"
Page = NoPage
__doc__ += apidoc(solara.lab.task)  # type: ignore
